"""Common GRPC backend client."""
import logging
import time
import warnings
from inspect import Parameter, Signature
from pathlib import Path
from types import ModuleType
from typing import (Any, Callable, ClassVar, Dict, Generator, Iterable, Iterator, Optional, Sequence, Type, Union, cast,
                    overload)

import grpc
from google.protobuf.empty_pb2 import Empty
from google.protobuf.message import Message
from grpc import ChannelCredentials, ssl_channel_credentials

from ._utils import get_call_id, lazy_format_message
from .constants import DEFAULT, DefaultType
from .decoder import GrpcDecoder
from .exceptions import ServiceNotFound

_LOGGER = logging.getLogger(__name__)

Request = Optional[Union[Message, Iterable[Message]]]


class CallWrapper:
    """Wrapper for gRPC calls.

    Arguments:
        call: A gRPC method to be called.
        request: A gRPC request to be sent.
        client: A GrpcClient instance.
        timeout: Timeout for the call.
        max_retries: A maximum number of calls to be made.
        modify_result: A callable to modify the result of the gRPC method.
    """

    def __init__(self, call: Callable, request: Request, *, client: 'GrpcClient', timeout: Optional[float] = None,
                 max_retries: int = 10, modify_result: Optional[Callable[[Any], Any]] = None):
        self.call = call
        self.request = request
        self.client = client
        self.timeout = timeout
        self.max_retries = max_retries
        self.modify_result = modify_result
        self._call_id = get_call_id()

    def get_reply(self) -> Any:
        """Fetch and return the reply."""
        return self.call(self.request, timeout=self.timeout)

    def decode_reply(self, reply: Message) -> Any:
        """Decode reply and return decoded result or raise decoded exception."""
        # Assert the supported structure of messages
        assert ('exception' in reply.DESCRIPTOR.fields_by_name or 'data' in reply.DESCRIPTOR.fields_by_name  # nosec
                or isinstance(reply, Empty))

        modify_result = self.modify_result or (lambda x: x)

        # Decode exception first
        if 'exception' in reply.DESCRIPTOR.fields_by_name and reply.HasField('exception'):
            raise self.client.decoder.decode_exception(reply.exception)  # type: ignore[attr-defined]

        # Decode data otherwise
        if 'data' in reply.DESCRIPTOR.fields_by_name and reply.HasField('data'):
            return modify_result(self.client.decoder.decode(reply.data))  # type: ignore[attr-defined]

        # The reply is empty
        return modify_result(None)

    def reconnect(self) -> None:
        """Try to reconnect the client if unavailable."""
        self.client.connect()

    def result(self) -> Any:
        """Return the call result."""
        # XXX: Extract the name of the gRPC method
        method = self.call._method.decode('utf-8')  # type: ignore
        _LOGGER.debug('[%s] %s(%s)', self._call_id, method, lazy_format_message(self.request))

        retries = 0
        try:
            while True:
                time.sleep(0.1 * retries)  # Wait for increasing amount of time before each call
                try:
                    retries += 1
                    reply = self.get_reply()
                    break
                except grpc.RpcError as error:
                    if error.code() == grpc.StatusCode.UNAVAILABLE and retries < self.max_retries:
                        # Connection probably failed. Try to reconnect.
                        _LOGGER.info('[%s] Connection failed. Trying to reconnect.', self._call_id)
                        self.reconnect()
                    else:
                        raise
        except grpc.RpcError as error:
            _LOGGER.info('[%s] %s failed with %s', self._call_id, method, error)
            raise

        _LOGGER.debug('[%s] %s returned %s', self._call_id, method, lazy_format_message(reply))

        return self.decode_reply(reply)


class StreamCallWrapper(CallWrapper):
    """Wrapper for gRPC call with streaming response."""

    def decode_reply(self, reply: Iterable[Message]) -> Generator[Any, None, None]:  # type: ignore[override]
        """Decode replies one by one."""
        for item in reply:
            yield super().decode_reply(item)


class FutureCallWrapper(CallWrapper):
    """Wrapper for gRPC future call."""

    def __init__(self, call: Callable, request: Request, *, client: 'GrpcClient', timeout: Optional[float] = None,
                 max_retries: int = 10, modify_result: Optional[Callable[[Any], Any]] = None):
        super().__init__(call, request, client=client, timeout=timeout, max_retries=max_retries,
                         modify_result=modify_result)
        # Get a future object - start an asynchronous request
        self._future = self._call_future()

    def _call_future(self) -> grpc.Future:
        """Start asynchronous request and return a future object."""
        future_call = self.call.future  # type: ignore
        return future_call(self.request, timeout=self.timeout)

    def get_reply(self) -> Any:
        """Fetch and return the reply."""
        return self._future.result()

    def reconnect(self) -> None:
        """Try to reconnect the client if unavailable."""
        # Cancel old request
        self._future.cancel()
        # Reconnect
        super().reconnect()
        # Start a new asynchronous request.
        self._future = self._call_future()

    def __getattr__(self, name: str) -> Any:
        """Proxy all other future properties."""
        return getattr(self._future, name)


class GrpcClient(object):
    """GRPC backend client.

    Class Attributes:
        decoder_cls: A default decoder class. No arguments are provided for its instantiation.
        grpc_modules: A default list of generated _pb2_grpc modules with service stubs.
            Deprecated in favor of `stub_cls`.
        options: Custom options for the gRPC client.
        stub_cls: The service stub for this client.
        timeout: A default timeout for gRPC calls. Default value is `None`, i.e. no timeout.

    Attributes:
        decoder: The decoder.
        grpc_modules: The list of generated _pb2_grpc modules with service stubs.
            Deprecated in favor of `stub_cls`.
        options: The final options of the gRPC client.
            The final options are merged from class and instance options, the later take precedence.
        stub_cls: The service stub for this client.
            If `stub_cls` is used, `service` argument can't be used in `call*` methods.
        timeout: Timeout for gRPC calls.
    """

    decoder_cls: ClassVar[Type[GrpcDecoder]] = GrpcDecoder
    grpc_modules: Sequence[ModuleType] = ()
    options: Dict[str, Any] = {}
    stub_cls: Optional[Type] = None
    timeout: Optional[float] = None

    def __init__(
        self,
        netloc: str,
        decoder: Optional[GrpcDecoder] = None,
        grpc_modules: Optional[Sequence[ModuleType]] = None,
        credentials: Optional[grpc.ChannelCredentials] = None,
        options: Optional[Dict[str, Any]] = None,
        timeout: Union[Optional[float], DefaultType] = DEFAULT,
        *,
        stub_cls: Optional[Type] = None,
    ):
        """Initialize GRPC backend client instance.

        Args:
            netloc: Network location of a gRPC server.
            decoder: Custom decoder instance.
            grpc_modules: Custom list of generated _pb2_grpc modules with service stubs.
                Deprecated in favor of `stub_cls`.
            credentials: Credentials for a secure channel connection. If None, insecure channel is used.
            options: Custom options for the gRPC client.
            stub_cls: Custom service stub for this client.
            timeout: Custom timeout for gRPC calls.
        """
        self.netloc = netloc
        self.credentials = credentials
        self.decoder = decoder or self.decoder_cls()
        self.grpc_modules = grpc_modules or self.grpc_modules
        if self.grpc_modules:
            warnings.warn("Argument 'grpc_modules' is deprecated, use 'stub_cls' instead.", DeprecationWarning,
                          stacklevel=2)
        self.options = dict(self.options, **(options or {}))
        self.stub_cls = stub_cls or self.stub_cls
        if timeout is not DEFAULT:
            self.timeout = timeout

        # Stubs cache
        self._stubs: Dict[str, Any] = {}

        self.connect()

    def connect(self) -> None:
        """Create channel to GPRC server."""
        if self.credentials is not None:
            self.channel = grpc.secure_channel(self.netloc, self.credentials, self.options.items())
        else:
            self.channel = grpc.insecure_channel(self.netloc, self.options.items())

    def get_stub(self, service: Optional[str] = None) -> Any:
        """Return stub for the service.

        Stubs are cached, so for each service, there is only one stub created.

        Arguments:
            service: Name of service, that is used to get service stub. Deprecated in favor of `stub_cls`.
        """
        if self.stub_cls is not None:
            stub_name = ''
            stub_cls = self.stub_cls
            if service is not None:
                warnings.warn("Argument 'service' is noop and deprecated.", DeprecationWarning, stacklevel=2)
        else:
            stub_name = '{}Stub'.format(service)
            for module in self.grpc_modules:
                if hasattr(module, stub_name):
                    stub_cls = getattr(module, stub_name)
                    break
            else:
                raise ServiceNotFound('Service {} could not be found'.format(stub_name))

        if stub_name in self._stubs:
            return self._stubs[stub_name]

        stub = stub_cls(self.channel)
        self._stubs[stub_name] = stub
        return stub

    _CALL_SIGNATURE = Signature([Parameter('method', Parameter.POSITIONAL_OR_KEYWORD),
                                 Parameter('request', Parameter.POSITIONAL_OR_KEYWORD),
                                 Parameter('timeout', Parameter.KEYWORD_ONLY, default=DEFAULT),
                                 Parameter('blocking', Parameter.KEYWORD_ONLY, default=None),
                                 Parameter('max_retries', Parameter.KEYWORD_ONLY, default=10),
                                 Parameter('modify_result', Parameter.KEYWORD_ONLY, default=None)])
    _CALL_SIGNATURE_OLD = Signature(
        [Parameter('service', Parameter.POSITIONAL_OR_KEYWORD)] + list(_CALL_SIGNATURE.parameters.values()))

    @overload
    def call(
        self, method: str, request: Request,
        *, timeout: Union[Optional[float], DefaultType] = DEFAULT, max_retries: int = 10,
        blocking: Optional[bool] = None, modify_result: Optional[Callable[[Any], Any]] = None,
    ) -> Any:
        ...

    @overload
    def call(
        self, service: str, method: str, request: Request,
        *, timeout: Union[Optional[float], DefaultType] = DEFAULT, max_retries: int = 10,
        blocking: Optional[bool] = None, modify_result: Optional[Callable[[Any], Any]] = None,
    ) -> Any:
        ...

    def call(self, *args: Any, **kwargs: Any) -> Any:
        """Call grpc method.

        Args:
            service: Name of service, that is used to get service stub. Deprecated in favor of `stub_cls`.
            method: Name of service method.
            request: Request passed as an argument to service method.
            timeout: An optional duration in seconds allowed for the gRPC call.
            blocking: Determines whether to use blocking or non-blocking call.
            max_retries: A maximum number of calls to be made.
            modify_result: Callable that processes decoded result and modifies it. Default is identity.
        """
        if self.stub_cls is not None:
            bound = self._CALL_SIGNATURE.bind(*args, **kwargs)
            bound.apply_defaults()
            service = None
        else:
            bound = self._CALL_SIGNATURE_OLD.bind(*args, **kwargs)
            bound.apply_defaults()
            service = bound.arguments['service']
        method = bound.arguments['method']
        request = bound.arguments['request']
        timeout = bound.arguments['timeout']
        blocking = bound.arguments['blocking']
        max_retries = bound.arguments['max_retries']
        modify_result = bound.arguments['modify_result']

        if timeout is DEFAULT:
            timeout = self.timeout
        if blocking is not None:
            if blocking:
                warnings.warn("Argument blocking is deprecated and will be deleted.", DeprecationWarning)
            else:
                warnings.warn("Argument blocking is deprecated, use method call_future instread.", DeprecationWarning)
                if self.stub_cls is not None:
                    return self.call_future(method, request, timeout=timeout, max_retries=max_retries,
                                            modify_result=modify_result)
                else:
                    return self.call_future(cast(str, service), method, request, timeout=timeout,
                                            max_retries=max_retries, modify_result=modify_result)
        stub = self.get_stub(service)
        grpc_method = getattr(stub, method)
        wrapper = CallWrapper(grpc_method, request, client=self, timeout=timeout, max_retries=max_retries,
                              modify_result=modify_result)
        return wrapper.result()

    _CALL_FUTURE_SIGNATURE = Signature([Parameter('method', Parameter.POSITIONAL_OR_KEYWORD),
                                        Parameter('request', Parameter.POSITIONAL_OR_KEYWORD),
                                        Parameter('timeout', Parameter.KEYWORD_ONLY, default=DEFAULT),
                                        Parameter('max_retries', Parameter.KEYWORD_ONLY, default=10),
                                        Parameter('modify_result', Parameter.KEYWORD_ONLY, default=None)])
    _CALL_FUTURE_SIGNATURE_OLD = Signature(
        [Parameter('service', Parameter.POSITIONAL_OR_KEYWORD)] + list(_CALL_SIGNATURE.parameters.values()))

    @overload
    def call_future(
        self, method: str, request: Request,
        *, timeout: Union[Optional[float], DefaultType] = DEFAULT, max_retries: int = 10,
        modify_result: Optional[Callable[[Any], Any]] = None,
    ) -> FutureCallWrapper:
        ...

    @overload
    def call_future(
        self, service: str, method: str, request: Request,
        *, timeout: Union[Optional[float], DefaultType] = DEFAULT, max_retries: int = 10,
        modify_result: Optional[Callable[[Any], Any]] = None,
    ) -> FutureCallWrapper:
        ...

    def call_future(self, *args: Any, **kwargs: Any) -> FutureCallWrapper:
        """Call grpc method asynchronously.

        Args:
            service: Name of service, that is used to get service stub. Deprecated in favor of `stub_cls`.
            method: Name of service method.
            request: Request passed as an argument to service method.
            timeout: An optional duration in seconds allowed for the gRPC call.
            max_retries: A maximum number of calls to be made.
            modify_result: Callable that processes decoded result and modifies it. Default is identity.
        """
        if self.stub_cls is not None:
            bound = self._CALL_FUTURE_SIGNATURE.bind(*args, **kwargs)
            bound.apply_defaults()
            service = None
        else:
            bound = self._CALL_FUTURE_SIGNATURE_OLD.bind(*args, **kwargs)
            bound.apply_defaults()
            service = bound.arguments['service']
        method = bound.arguments['method']
        request = bound.arguments['request']
        timeout = bound.arguments['timeout']
        max_retries = bound.arguments['max_retries']
        modify_result = bound.arguments['modify_result']

        if timeout is DEFAULT:
            timeout = self.timeout
        stub = self.get_stub(service)
        grpc_method = getattr(stub, method)
        return FutureCallWrapper(grpc_method, request, client=self, timeout=timeout, max_retries=max_retries,
                                 modify_result=modify_result)

    @overload
    def call_stream(
        self, method: str, request: Request,
        *, timeout: Union[Optional[float], DefaultType] = DEFAULT, max_retries: int = 10,
        modify_result: Optional[Callable[[Any], Any]] = None,
    ) -> Iterator[Any]:
        ...

    @overload
    def call_stream(
        self, service: str, method: str, request: Request,
        *, timeout: Union[Optional[float], DefaultType] = DEFAULT, max_retries: int = 10,
        modify_result: Optional[Callable[[Any], Any]] = None,
    ) -> Iterator[Any]:
        ...

    def call_stream(self, *args: Any, **kwargs: Any) -> Iterator[Any]:
        """Call grpc method with streaming response.

        Args:
            service: Name of service, that is used to get service stub. Deprecated in favor of `stub_cls`.
            method: Name of service method.
            request: Request passed as an argument to service method.
            timeout: An optional duration in seconds allowed for the gRPC call.
            max_retries: A maximum number of calls to be made.
            modify_result: Callable that processes decoded result and modifies it. Default is identity.
        """
        # This method has the same signature as call_future, so reuse the definitions.
        if self.stub_cls is not None:
            bound = self._CALL_FUTURE_SIGNATURE.bind(*args, **kwargs)
            bound.apply_defaults()
            service = None
        else:
            bound = self._CALL_FUTURE_SIGNATURE_OLD.bind(*args, **kwargs)
            bound.apply_defaults()
            service = bound.arguments['service']
        method = bound.arguments['method']
        request = bound.arguments['request']
        timeout = bound.arguments['timeout']
        max_retries = bound.arguments['max_retries']
        modify_result = bound.arguments['modify_result']

        if timeout is DEFAULT:
            timeout = self.timeout
        stub = self.get_stub(service)
        grpc_method = getattr(stub, method)
        wrapper = StreamCallWrapper(grpc_method, request, client=self, timeout=timeout, max_retries=max_retries,
                                    modify_result=modify_result)
        return wrapper.result()  # type: ignore[no-any-return]


def make_credentials(ssl_cert: Optional[Union[str, Path]]) -> Optional[ChannelCredentials]:
    """Create client-side credentials from SSL certificate file.

    Args:
        ssl_cert: Path to a SSL certificate file.

    Returns:
        None if ssl_cert is None.
        Credentials otherwise.

    Raises:
        OSError: If an error on SSL certificate file is encountered.
    """
    if ssl_cert is not None:
        with open(ssl_cert) as file:
            return ssl_channel_credentials(file.read())
    else:
        return None
