from unittest import TestCase
from unittest.mock import patch

from frgal._utils import MESSAGE_MAX_LENGTH, format_message, get_call_id, lazy_format_message

from .proto.example_pb2 import SimpleMessage  # type: ignore[attr-defined]


class GetCallIdTest(TestCase):
    def test_get_call_id(self):
        call_id = get_call_id()

        self.assertIsInstance(call_id, str)
        self.assertEqual(len(call_id), 4)

    def test_random(self):
        self.assertNotEqual(get_call_id(), get_call_id())


class FormatMessageTest(TestCase):
    def test_short_message(self):
        message = SimpleMessage()
        message.value = "I am fish!"
        self.assertEqual(format_message(message), 'value: "I am fish!"')

    def test_long_message(self):
        message = SimpleMessage()
        message.value = "I am fish! " * 400

        out = format_message(message)

        self.assertEqual(len(out), MESSAGE_MAX_LENGTH)
        self.assertRegex(out, r'^value: "I am fish! .*\.\.\.$')

    def test_short_object(self):
        self.assertEqual(format_message(42), '42')

    def test_long_object(self):
        message = ["I am fish!"] * 400
        self.assertRegex(format_message(message), r"^\['I am fish!', .*\.\.\.$")

    def test_generator(self):
        obj = ("I am fish!" for i in range(400))  # pragma: no cover
        self.assertRegex(format_message(obj), r'^<generator object .*>$')


class LazyFormatMessageTest(TestCase):
    def test(self):
        # Test evaluation of the lazy wrapper.
        data = (
            (SimpleMessage(value="I am fish!"), 'value: "I am fish!"'),
            (42, '42'),
        )
        for message, result in data:
            with self.subTest(message=message):
                self.assertEqual(str(lazy_format_message(message)), result)
                self.assertEqual(repr(lazy_format_message(message)), result)
                self.assertEqual('{}'.format(lazy_format_message(message)), result)
                self.assertEqual('%s' % lazy_format_message(message), result)
                self.assertEqual('%r' % lazy_format_message(message), result)

    def test_shorten(self):
        # Test message shortening
        data = (
            (SimpleMessage(value="I am fish!" * 400), 21, 'value: "I am fish!...'),
            (["I am fish!"] * 400, 18, "['I am fish!', ..."),
        )
        for message, max_length, result in data:
            with self.subTest(message=message, max_length=max_length):
                with patch('frgal._utils.MESSAGE_MAX_LENGTH', new=max_length):
                    self.assertEqual(str(lazy_format_message(message)), result)
                    self.assertEqual(repr(lazy_format_message(message)), result)
                    self.assertEqual('{}'.format(lazy_format_message(message)), result)
                    self.assertEqual('%s' % lazy_format_message(message), result)
                    self.assertEqual('%r' % lazy_format_message(message), result)
